import tensorflow as tf

@tf.keras.utils.register_keras_serializable(package="physDBD")
def unit_mat_sym(n: int, i: int, j: int):
    """Construct the symmetric unit matrix of size nxn
       1 at (i,j) AND (j,i)
       0 elsewhere

    Args:
        n (int): Size of square matrix
        i (int): First idx
        j (int): Second idx

    Returns:
        tf.Constant: Matrix that is 1 at (i,j) AND (j,i) and 0 everywhere else
    """
    idx = i * n + j
    one_hot = tf.one_hot(indices=idx,depth=n*n, dtype='float32')
    
    if i != j:
        idx = j * n + i
        one_hot += tf.one_hot(indices=idx,depth=n*n, dtype='float32')

    return tf.reshape(one_hot,shape=(n,n))

@tf.keras.utils.register_keras_serializable(package="physDBD")
def unit_mat(n: int, i: int, j: int):
    """Construct the non-symmetric unit matrix of size nxn
       1 at (i,j) ONLY
       0 elsewhere

    Args:
        n (int): Size of square matrix
        i (int): First idx
        j (int): Second idx

    Returns:
        tf.Constant: Matrix that is 1 at (i,j) ONLY and 0 everywhere else
    """
    idx = i * n + j
    one_hot = tf.one_hot(indices=idx,depth=n*n, dtype='float32')
    return tf.reshape(one_hot,shape=(n,n))


from ..net_common import unit_mat_sym
from ..helpers import check_non_zero_idx_pairs

import tensorflow as tf

import numpy as np

from typing import List, Tuple

@tf.keras.utils.register_keras_serializable(package="tfConstrainedGauss")
class LayerMultPrecCov(tf.keras.layers.Layer):
    """Layer that multiplies precision by covarinace matrix
    """

    @classmethod
    def constructDiag(cls,
        n: int,
        non_zero_idx_pairs: List[Tuple[int,int]],
        init_diag_val: float = 1.0,
        **kwargs
        ):
        """Constructor for a diagonal precision matrix with elements specified

        Args:
            n (int): Size of matrix
            non_zero_idx_pairs (List[Tuple[int,int]]): List of non-zero index pairs in the matrix
            init_diag_val (float, optional): Value of the diagonal part of precision matrix. Defaults to 1.0.

        Returns:
            LayerMultPrecCov: layer
        """
        check_non_zero_idx_pairs(n, non_zero_idx_pairs)

        # Set all diagonal elements to one, rest zero (like a identity matrix)
        non_zero_vals = np.zeros(len(non_zero_idx_pairs))
        for i,pair in enumerate(non_zero_idx_pairs):
            if pair[0] == pair[1]:
                non_zero_vals[i] = init_diag_val

        return cls(
            n=n,
            non_zero_idx_pairs=non_zero_idx_pairs,
            non_zero_vals=non_zero_vals,
            **kwargs
            )

    def __init__(self,
        n: int,
        non_zero_idx_pairs: List[Tuple[int,int]],
        non_zero_vals: np.array,
        **kwargs
        ):
        """Constructor

        Args:
            n (int): Size of matrix
            non_zero_idx_pairs (List[Tuple[int,int]]): List of non-zero index pairs in the matrix
            non_zero_vals (np.array): Non-zero values in the precision matrix corresponding to non_zero_idx_pairs
        """
        super(LayerMultPrecCov, self).__init__(**kwargs)

        check_non_zero_idx_pairs(n, non_zero_idx_pairs)

        self.n = n
        self.non_zero_idx_pairs = non_zero_idx_pairs

        self.non_zero_vals = self.add_weight(
            name="non_zero_vals",
            shape=len(non_zero_vals),
            trainable=True,
            initializer=tf.constant_initializer(non_zero_vals),
            dtype='float32'
            )

    @property
    def n_non_zero(self) -> int:
        """Number of non-zero elements, not counting twice for symmetry!

        Returns:
            int: number of non-zero elements, not counting twice for symmetry!
        """
        return len(self.non_zero_idx_pairs)

    def get_config(self):
        """Get config for writing
        """
        config = super(LayerMultPrecCov, self).get_config()
        config.update({
            "n": self.n,
            "non_zero_idx_pairs": self.non_zero_idx_pairs,
            "non_zero_vals": self.non_zero_vals.numpy()
            })
        return config

    @classmethod
    def from_config(cls, config):
        """Construct from config
        """
        return cls(**config)

    def call(self, inputs):
        """Call the layer

        Args:
            inputs ([type]): Covariance matrix: batch_size x n x n

        Returns:
            [type]: Product of precision and covariance matrix: vector of length batch_size
        """
        # Inputs = cov mat

        prec_mat = tf.zeros_like(inputs)
        for i,pair in enumerate(self.non_zero_idx_pairs):
            prec_mat_0 = self.non_zero_vals[i] * unit_mat_sym(self.n,pair[0],pair[1])
            prec_mat += tf.map_fn(lambda val: prec_mat_0, prec_mat)

        return tf.matmul(prec_mat,inputs)

@tf.keras.utils.register_keras_serializable(package="tfConstrainedGauss")
class ModelID(tf.keras.Model):
    """Model for the Identidy method
    """

    def __init__(self, 
        mult_lyr: LayerMultPrecCov,
        **kwargs
        ):
        super(ModelID, self).__init__(**kwargs)

        self.mult_lyr = mult_lyr

    def get_config(self):
        # Do not call super.get_config !!!
        config = {
            "mult_lyr": self.mult_lyr
        }
        return config

    # from_config doesn't get called anyways?
    @classmethod
    def from_config(cls, config):
        return cls(**config)

    def call(self, input_tensor, training=False):
        """Wrapper to call the mulitplication layer of precision and covariance matrices
        
        Args:
            input_tensor ([type]): Covariance matrix: batch_size x n x n

        Returns:
            [type]: Product of precision and covariance matrix: vector of length batch_size
        """
        return self.mult_lyr(input_tensor)


from ..net_common import unit_mat_sym
from ..helpers import check_non_zero_idx_pairs

import tensorflow as tf
import numpy as np

from typing import List, Tuple

@tf.keras.utils.register_keras_serializable(package="tfConstrainedGauss")
class LayerPrecToCovMat(tf.keras.layers.Layer):
    """Layer that converts precision to covariance matrix
    """

    @classmethod
    def constructDiag(cls,
        n: int,
        non_zero_idx_pairs: List[Tuple[int,int]],
        init_diag_val: float = 1.0,
        **kwargs
        ):
        """Constructor with initial precision matrix that is diagonsl

        Args:
            n (int): Size of matirx
            non_zero_idx_pairs (List[Tuple[int,int]]): List of lower triangular index pairs for non-zero elements in precision matrix
            init_diag_val (float, optional): Initial diagonal value of precision matrix. Defaults to 1.0.

        Returns:
            LayerPrecToCovMat: layer
        """
        check_non_zero_idx_pairs(n, non_zero_idx_pairs)

        # Set all diagonal elements to one, rest zero (like a identity matrix)
        non_zero_vals = np.zeros(len(non_zero_idx_pairs))
        for i,pair in enumerate(non_zero_idx_pairs):
            if pair[0] == pair[1]:
                non_zero_vals[i] = init_diag_val

        return cls(
            n=n,
            non_zero_idx_pairs=non_zero_idx_pairs,
            non_zero_vals=non_zero_vals,
            **kwargs
            )

    def __init__(self,
        n: int,
        non_zero_idx_pairs: List[Tuple[int,int]],
        non_zero_vals: np.array,
        **kwargs
        ):
        """Constructor

        Args:
            n (int): Size of matirx
            non_zero_idx_pairs (List[Tuple[int,int]]): List of lower triangular index pairs for non-zero elements in precision matrix
            non_zero_vals (np.array): Non-zero values corresponding to non_zero_idx_pairs
        """
        super(LayerPrecToCovMat, self).__init__(**kwargs)

        check_non_zero_idx_pairs(n, non_zero_idx_pairs)

        self.n = n
        self.non_zero_idx_pairs = non_zero_idx_pairs

        self.non_zero_vals = self.add_weight(
            name="non_zero_vals",
            shape=len(non_zero_vals),
            trainable=True,
            initializer=tf.constant_initializer(non_zero_vals),
            dtype='float32'
            )

    @property
    def n_non_zero(self):
        """Number of non-zero unique elements in precision matrix (excluding symmetry)

        Returns:
            int: Count
        """
        return len(self.non_zero_idx_pairs)

    def get_config(self):
        config = super(LayerPrecToCovMat, self).get_config()
        config.update({
            "n": self.n,
            "non_zero_idx_pairs": self.non_zero_idx_pairs,
            "non_zero_vals": self.non_zero_vals.numpy()
            })
        return config

    @classmethod
    def from_config(cls, config):
        return cls(**config)

    def call(self, inputs):
        """Construct the full precision matrix from the non-zero values and invert it to return the covariance matrix

        Args:
            inputs ([type]): Anything; used only to determine batch size: size should be batch_size x whatever

        Returns:
            [type]: Covariance matrix: batch_size x n x n
        """
        # Form precision matrix
        batch_size = tf.shape(inputs)[0]
        prec_mat = tf.zeros((batch_size,self.n,self.n), dtype='float32')

        for i,pair in enumerate(self.non_zero_idx_pairs):
            prec_mat += tf.map_fn(lambda val: self.non_zero_vals[i] * unit_mat_sym(self.n,pair[0],pair[1]), prec_mat)
        
        # Invert to get Covariance matrix
        cov_mat = tf.linalg.inv(prec_mat)

        return cov_mat

@tf.keras.utils.register_keras_serializable(package="tfConstrainedGauss")
class ModelME(tf.keras.Model):
    """Model for the MaxEnt method
    """

    def __init__(self, 
        inv_lyr: LayerPrecToCovMat,
        **kwargs
        ):
        """Constructor

        Args:
            inv_lyr (LayerPrecToCovMat): Precision to covariance matrix layer
        """
        super(ModelME, self).__init__(**kwargs)

        self.inv_lyr = inv_lyr

    def get_config(self):
        # Do not call super.get_config !!!
        config = {
            "inv_lyr": self.inv_lyr
        }
        return config

    # from_config doesn't get called anyways?
    @classmethod
    def from_config(cls, config):
        return cls(**config)

    def call(self, input_tensor, training=False):
        """Given some input_tensor that is only used to set the batch size (must be batch_size >= 2, else there are problems!)
            Calls the precision matrix to covariance matrix layer and returns the vector of the elements in the covariance matrix
            corresponding to the non-zero entries in the precision matrix.

        Args:
            input_tensor ([type]): Anything; used only to determine batch size: size should be batch_size x whatever

        Returns:
            [type]: Elements in the covariance matrix corresponding to the non-zero entries in the precision matrix.: batch_size x no_non_zero
        """
        cov_mat = self.inv_lyr(input_tensor)

        batch_size = tf.shape(cov_mat)[0]
        cov_mat_non_zero = tf.zeros((batch_size,self.inv_lyr.n_non_zero), dtype='float32')

        for i,pair in enumerate(self.inv_lyr.non_zero_idx_pairs):
            unit_vec = tf.one_hot(
                indices=i,
                depth=self.inv_lyr.n_non_zero,
                dtype='float32'
                )
            cov_mat_non_zero += tf.map_fn(lambda cov_mat_batch: cov_mat_batch[pair[0],pair[1]] * unit_vec, cov_mat)

        return cov_mat_non_zero


from ..helpers import convert_mat_non_zero_to_inv_mat, check_symmetric, \
    convert_mat_non_zero_to_inv_mat_non_zero, convert_mat_non_zero_to_mat, \
        convert_mat_to_mat_non_zero
from .model_id import ModelID, LayerMultPrecCov

import tensorflow as tf

import numpy as np

from typing import List, Tuple
from dataclasses import dataclass

@dataclass
class InputsID:
    n : int
    non_zero_idx_pairs: List[Tuple[int,int]]
    target_cov_mat : np.array
    epochs: int = 100
    learning_rate: float = 0.01

    def convert_mat_non_zero_to_inv_mat_non_zero(self, mat_non_zero: np.array) -> np.array:
        return convert_mat_non_zero_to_inv_mat_non_zero(
            n=self.n,
            non_zero_idx_pairs=self.non_zero_idx_pairs,
            mat_non_zero=mat_non_zero
        )
    
    def convert_mat_non_zero_to_inv_mat(self, mat_non_zero: np.array) -> np.array:
        return convert_mat_non_zero_to_inv_mat(
            n=self.n,
            non_zero_idx_pairs=self.non_zero_idx_pairs,
            mat_non_zero=mat_non_zero
        )

    def convert_mat_to_mat_non_zero(self, mat: np.array) -> np.array:
        return convert_mat_to_mat_non_zero(
            n=self.n,
            non_zero_idx_pairs=self.non_zero_idx_pairs,
            mat=mat
        )
    
    @property
    def n_non_zero(self):
        return len(self.non_zero_idx_pairs)

    def report(self):
        print("----- Inputs -----")

        print("Constraints on the prec mat: non-zero elements are:")
        print(self.non_zero_idx_pairs)

        print("Target cov mat:")
        print(self.target_cov_mat)

        print("Solver options:")
        print("epochs:", self.epochs)
        print("learning_rate:", self.learning_rate)

        print("----- End inputs -----")

@dataclass
class ResultsID:
    inputs: InputsID
    trained_model : ModelID

    init_prec_mat_non_zero : np.array
    init_cov_mat_reconstructed : np.array
    learned_prec_mat_non_zero : np.array
    learned_cov_mat : np.array

    @property
    def learned_prec_mat(self):
        return convert_mat_non_zero_to_mat(
            n=self.inputs.n,
            non_zero_idx_pairs=self.inputs.non_zero_idx_pairs,
            mat_non_zero=self.learned_prec_mat_non_zero
            )

    def report(self):
        print("----- Results -----")

        print("Prec mat initial guess for non-zero elements:")
        print(self.init_prec_mat_non_zero)

        print("-> Learned prec mat non-zero elements:")
        print(self.learned_prec_mat_non_zero)

        print("Initial cov mat corresponding to initial prec mat guess:")
        print(self.init_cov_mat_reconstructed)

        print("-> Learned cov mat:")
        print(self.learned_cov_mat)

        print("--> Target cov mat:")
        print(self.inputs.target_cov_mat)

        print("Matrix product (prec.cov) learned:")
        print(np.dot(self.learned_prec_mat,self.learned_cov_mat))

        print("----- End results -----")

def solve_id(
    inputs: InputsID
    ) -> ResultsID:
    """Solve the identity proble

    Args:
        inputs (InputsID): Inputs

    Returns:
        ResultsID: Results
    """

    assert(inputs.target_cov_mat.shape == (inputs.n,inputs.n))
    assert(check_symmetric(inputs.target_cov_mat))
    
    # Invert
    init_prec_mat = np.linalg.inv(inputs.target_cov_mat)
    init_prec_mat_non_zero = inputs.convert_mat_to_mat_non_zero(init_prec_mat)
    init_cov_mat_reconstructed = inputs.convert_mat_non_zero_to_inv_mat(init_prec_mat_non_zero)

    lyr = LayerMultPrecCov(
        n=inputs.n,
        non_zero_idx_pairs=inputs.non_zero_idx_pairs,
        non_zero_vals=init_prec_mat_non_zero
        )

    model = ModelID(mult_lyr=lyr)

    loss_fn = tf.keras.losses.MeanSquaredError()
    opt = tf.keras.optimizers.Adam(learning_rate=inputs.learning_rate)
    model.compile(optimizer=opt,
                loss=loss_fn,
                run_eagerly=False)

    # Covariance matrix input
    # NOTE: do not use batch size = 1, anything higher eg 2 is OK - 1 is peculiar
    batch_size = 2
    train_inputs = np.full(shape=(batch_size,inputs.n,inputs.n),fill_value=inputs.target_cov_mat)

    # Output = identity
    train_outputs = np.full(shape=(batch_size,inputs.n,inputs.n),fill_value=np.eye(inputs.n))

    # Train!
    # Do NOT pass batch_size = 1 -> peculiar problems
    tf.get_logger().setLevel('ERROR')
    model.fit(
        train_inputs, 
        train_outputs, 
        epochs=inputs.epochs, 
        batch_size=batch_size
        )

    # Return
    learned_prec_mat_non_zero = model.mult_lyr.non_zero_vals.numpy()
    learned_cov_mat = inputs.convert_mat_non_zero_to_inv_mat(learned_prec_mat_non_zero)

    return ResultsID(
        inputs=inputs,
        trained_model=model,
        init_prec_mat_non_zero=init_prec_mat_non_zero,
        init_cov_mat_reconstructed=init_cov_mat_reconstructed,
        learned_prec_mat_non_zero=learned_prec_mat_non_zero,
        learned_cov_mat=learned_cov_mat
        )


import numpy as np
import pandas as pd
from dataclasses import astuple
from typing import List, Tuple

def random_non_zero_idx_pairs(n: int) -> List[Tuple[int,int]]:
    """Generate a random list of non-zero index pairs. The number of index pairs is less than n choose 2 (max).
    The indices are lower triangular.

    Args:
        n (int): Size of the matrix

    Returns:
        List[Tuple[int,int]]: List of index pairs in the lower triangular matrix.
    """
    
    # Non zero elements
    non_zero_idx_pairs = []
    
    # All diagonal (required)
    for i in range(0,n):
        non_zero_idx_pairs.append((i,i))
    
    # Some off diagonal < n choose 2
    max_no_off_diagonal = int((n-1)*n/2)
    no_off_diagonal = np.random.randint(low=0,high=max_no_off_diagonal)
    print("No non-zero off-diagonal elements:",no_off_diagonal,"max possible:",max_no_off_diagonal)
    
    idx = 0
    while idx < no_off_diagonal:
        i = np.random.randint(low=1,high=n)
        j = np.random.randint(low=0,high=i)
        if not (i,j) in non_zero_idx_pairs:
            non_zero_idx_pairs.append((i,j))
            idx += 1

    return non_zero_idx_pairs

# Random realistic cov mat following Madar 2015 
# Stat Probab Lett. 2015 August 1; 103: 142–147. doi:10.1016/j.spl.2015.03.014.
def random_cov_mat(n: int, unit_diag: bool, diag_scale_std_dev: float = 0.1) -> np.array:
    """Generate a random realistic covariance matrix following
    Madar 2015
    Stat Probab Lett. 2015 August 1; 103: 142–147. doi:10.1016/j.spl.2015.03.014.

    Args:
        n (int): Size of matrix
        unit_diag (bool): Whether the diagonal should be all 1 (True), otherwise the diagonal is scaled as 
            described by diag_scale_std_dev
        diag_scale_std_dev (float): If unit_diag is False, a scaling matrix is constructed as 
            A = diag(randomNormal(mean=1, std_dev=diag_scale_std_dev)). The covariance matrix is transformed as A.cov.Transpose[A]
            corresponding to scaling the data matrix Z by A.Z. Default = 0.1.

    Returns:
        np.array: Covariance matrix
    """
    
    # Step 1
    u = np.random.rand(n)
    
    # Sort descending
    u = -np.sort(-u)

    l = np.zeros(shape=(n,n))
    l[0,0] = 1.0
    l[1,1] = np.sqrt(u[1])
    for j in range(2,n):
        l[j,j] = np.sqrt(u[j]/u[j-1])

    # Step 2
    um = np.zeros(shape=(n,n))
    # Original idx: 2 to n-1 inclusive = 2 to n
    # Zero idx: 1 to n
    for j in range(1,n):
        um[0,j] = 1.0
        um[j,j] = pow(l[j,j],2)

        # No additional random variables = j-2 (original)
        # So j-1 in zero indexed version
        if j-1 > 0:
            # Random decreasing
            um_extra = np.random.uniform(low=l[j,j],high=1.0,size=(j-1))
            um_extra = - np.sort(-um_extra)

            # Assign
            um[1:j,j] = um_extra

        # Original idx: 1 to j-1 inclusive = 1 to j
        # Zero idx: 0 to j
        for i in range(0,j):
            l[j,i] = np.sqrt(um[i,j] - um[i+1,j])

    # Step 3
    for i in range(0,n):
        for j in range(i+1,n):
            bij = np.random.binomial(n=n,p=0.5)
            l[j,i] *= pow(-1,bij)
    
    # Step 4
    cov_mat = np.dot(l, np.transpose(l))

    # Scaling (optional)
    if not unit_diag:

        # Scaling matrix
        a_diag = np.random.normal(
            loc=1.0, 
            scale=diag_scale_std_dev, 
            size=n
            )
        a = np.diag(a_diag)

        # Scale data matrix z linearly by a.z
        # Covariance scales by cov(a.z) = a.cov(z).transpose(a)
        cov_mat = np.dot(a, np.dot(cov_mat, np.transpose(a)))

    return cov_mat

def check_symmetric(a, rtol=1e-05, atol=1e-08) -> bool:
    """Check a matrix is symmetric

    Args:
        a ([type]): Matrix
        rtol ([type], optional): Tol. Defaults to 1e-05.
        atol ([type], optional): Tol. Defaults to 1e-08.

    Returns:
        bool: True if symmetric
    """
    return np.allclose(a, a.T, rtol=rtol, atol=atol)

def convert_mat_non_zero_to_mat(n: int, non_zero_idx_pairs: List[Tuple[int,int]], mat_non_zero: np.array) -> np.array:
    """Convert list of matrice's non-zero elements into the matrix

    Args:
        n (int): Size of matrix
        non_zero_idx_pairs (List[Tuple[int,int]]): List of non-zero index pairs in the matrix
        mat_non_zero (np.array): Corresponding non-zero matrix elements

    Returns:
        np.array: nxn matrix with specified non-zero elements
    """
    mat = np.zeros((n,n))
    for i,pair in enumerate(non_zero_idx_pairs):
        mat[pair[0],pair[1]] = mat_non_zero[i]
        mat[pair[1],pair[0]] = mat_non_zero[i]
    return mat

def convert_mat_non_zero_to_inv_mat(n: int, non_zero_idx_pairs: List[Tuple[int,int]], mat_non_zero: np.array) -> np.array:
    """Convert list of matrice's non-zero elements into the matrice's inverse 

    Args:
        n (int): Size of matrix
        non_zero_idx_pairs (List[Tuple[int,int]]): List of non-zero index pairs in the matrix
        mat_non_zero (np.array): Corresponding non-zero matrix elements

    Returns:
        np.array: Inverse of the nxn matrix
    """
    mat = convert_mat_non_zero_to_mat(
        n=n,
        non_zero_idx_pairs=non_zero_idx_pairs,
        mat_non_zero=mat_non_zero
        )
    inv_mat = np.linalg.inv(mat)
    return inv_mat

def convert_mat_non_zero_to_inv_mat_non_zero(n: int, non_zero_idx_pairs: List[Tuple[int,int]], mat_non_zero: np.array) -> np.array:
    """Convert list of matrice's non-zero elements into the matrice's inverse's list of elements

    Args:
        n (int): Size of matrix
        non_zero_idx_pairs (List[Tuple[int,int]]): List of non-zero index pairs in the matrix
        mat_non_zero (np.array): Corresponding non-zero matrix elements

    Returns:
        np.array: List of elements corresponding to indices non_zero_idx_pairs in the inverse of the matrix which has non-zero elements given by mat_non_zero
    """
    inv_mat = convert_mat_non_zero_to_inv_mat(
        n=n,
        non_zero_idx_pairs=non_zero_idx_pairs,
        mat_non_zero=mat_non_zero
        )
    inv_mat_non_zero = convert_mat_to_mat_non_zero(
        n=n,
        non_zero_idx_pairs=non_zero_idx_pairs,
        mat=inv_mat
        )
    return inv_mat_non_zero

def convert_mat_to_mat_non_zero(n: int, non_zero_idx_pairs: List[Tuple[int,int]], mat: np.array) -> np.array:
    """Convert a matrix to it's list of non-zero elements for specified non-zero indices

    Args:
        n (int): Size of matrix
        non_zero_idx_pairs (List[Tuple[int,int]]): List of non-zero index pairs in the matrix
        mat_non_zero (np.array): nxn matrix
    
    Returns:
        np.array: List of elements specified by non_zero_idx_pairs
    """
    assert(check_symmetric(mat))

    mat_non_zero = np.zeros(len(non_zero_idx_pairs))
    for i,pair in enumerate(non_zero_idx_pairs):
        mat_non_zero[i] = mat[pair[0],pair[1]]
    return mat_non_zero

def check_non_zero_idx_pairs(n: int, non_zero_idx_pairs: List[Tuple[int,int]]):
    """Check list of non-zero index pairs is valid, i.e. lower triangular and that all diagonal elements are given

    Args:
        n (int): Size of matrix
        non_zero_idx_pairs (List[Tuple[int,int]]): List of non-zero index pairs in the matrix

    Raises:
        ValueError: If not all diagonal elements are specified
        ValueError: If some indices are not lower-triangular
    """
    for i in range(0,n):
        if not (i,i) in non_zero_idx_pairs:
            raise ValueError("All diagonal elements must be specified as non-zero.")

    for pair in non_zero_idx_pairs:
        if pair[0] < pair[1]:
            raise ValueError("Only provide lower triangular indexes.")

def convert_np_to_pd(arr_with_times: np.array, nv: int, nh: int) -> pd.DataFrame:
    """Convert a numpy array of wt, b, sig2, muh, varh_diag to a pandas dataframe with named columns

    Args:
        arr_with_times (np.array): Array of size TxN where T is the number of timepoints and 
            N is the size of (wt,b,sig2,muh,varh_diag) = (nv*nh, nv, 1, nh, nh) = nv*nh + nv + 1 + 2*nh
        nv (int): No. visible species
        nh (int): No. hidden species

    Returns:
        pd.DataFrame: Pandas data frame
    """

    # Convert to pandas
    columns = ["t"]
    for ih in range(0,nh):
        for iv in range(0,nv):
            columns += ["wt%d%d" % (ih,iv)]
    for iv in range(0,nv):
        columns += ["b%d" % iv]
    columns += ["sig2"]
    for ih in range(0,nh):
        columns += ["muh%d" % ih]
    for ih in range(0,nh):
        columns += ["varh_diag%d" % ih]

    df = pd.DataFrame(arr_with_times, columns=columns)
    return df

def normalize(vec: np.array) -> np.array:
    """Normalize 1D np arr

    Args:
        vec (np.array): 1D vec to normalize

    Returns:
        np.array: normalized
    """
    return vec / np.sqrt(np.sum(vec**2))

# Equality of np arrays in data class
# https://stackoverflow.com/a/51743960/1427316
def array_safe_eq(a, b) -> bool:
    """Check if a and b are equal, even if they are numpy arrays. 
        Needed for checking equality with @dataclass(eq=False) decorator"""
    if a is b:
        return True
    if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
        return a.shape == b.shape and np.max(abs(a - b)) < 1e-8
    if isinstance(a,float) and isinstance(b,float):
        return np.max(abs(a-b)) < 1e-8
    try:
        return a == b
    except TypeError:
        return NotImplemented

def dc_eq(dc1, dc2) -> bool:
   """checks if two dataclasses which hold numpy arrays are equal. 
    Needed for checking equality with @dataclass(eq=False) decorator"""
   if dc1 is dc2:
        return True
   if dc1.__class__ is not dc2.__class__:
       return NotImplemented  # better than False
   t1 = astuple(dc1)
   t2 = astuple(dc2)
   return all(array_safe_eq(a1, a2) for a1, a2 in zip(t1, t2))


from ..helpers import convert_mat_to_mat_non_zero, convert_mat_non_zero_to_mat, \
    convert_mat_non_zero_to_inv_mat_non_zero, convert_mat_non_zero_to_inv_mat
from .model_me import ModelME, LayerPrecToCovMat

import tensorflow as tf
import numpy as np
from dataclasses import dataclass

from typing import List, Tuple

@dataclass
class InputsME:
    """Inputs to the MaxEnt problem
    """

    n : int
    non_zero_idx_pairs: List[Tuple[int,int]]
    target_cov_mat_non_zero : np.array
    epochs: int = 100
    learning_rate: float = 0.01
    use_weighted_loss : bool = False

    verbose : bool = False

    def convert_mat_non_zero_to_inv_mat_non_zero(self, mat_non_zero: np.array):
        """Wrapper around convert_mat_non_zero_to_inv_mat_non_zero
        """
        return convert_mat_non_zero_to_inv_mat_non_zero(
            n=self.n,
            non_zero_idx_pairs=self.non_zero_idx_pairs,
            mat_non_zero=mat_non_zero
        )
    
    @property
    def n_non_zero(self):
        """The number of unique non-zero elements in the precision matrix (excluding symmetry)

        Returns:
            int: Count
        """
        return len(self.non_zero_idx_pairs)

    def report(self):
        """Print some information
        """
        print("----- Inputs -----")

        print("Constraints on the prec mat: non-zero elements are:")
        print(self.non_zero_idx_pairs)

        print("Constraints on the cov mat: specified values corresponding to above idxs:")
        print(self.target_cov_mat_non_zero)

        print("Solver options:")
        print("epochs:", self.epochs)
        print("learning_rate:", self.learning_rate)
        print("use_weighted_loss:", self.use_weighted_loss)

        print("----- End inputs -----")

@dataclass
class ResultsME:
    """Results of the MaxEnt problem
    """
    inputs: InputsME
    trained_model : ModelME

    init_prec_mat_non_zero : np.array
    init_cov_mat_reconstructed_non_zero : np.array
    learned_prec_mat_non_zero : np.array
    learned_cov_mat : np.array

    @property
    def learned_prec_mat(self) -> np.array:
        return convert_mat_non_zero_to_mat(
            n=self.inputs.n,
            non_zero_idx_pairs=self.inputs.non_zero_idx_pairs,
            mat_non_zero=self.learned_prec_mat_non_zero
            )

    @property
    def learned_cov_mat_non_zero(self) -> np.array:
        return convert_mat_to_mat_non_zero(
            n=self.inputs.n,
            non_zero_idx_pairs=self.inputs.non_zero_idx_pairs,
            mat=self.learned_cov_mat
            )

    def report(self):
        """Print some information
        """
        print("----- Results -----")

        print("Prec mat initial guess for non-zero elements:")
        print(self.init_prec_mat_non_zero)

        print("-> Learned prec mat non-zero elements:")
        print(self.learned_prec_mat_non_zero)

        print("Initial cov mat non-zero elements corresponding to initial prec mat guess:")
        print(self.init_cov_mat_reconstructed_non_zero)

        print("-> Learned cov mat non-zero elements:")
        print(self.learned_cov_mat_non_zero)

        print("--> Target cov mat non-zero elements:")
        print(self.inputs.target_cov_mat_non_zero)

        print("----- End results -----")

def custom_weighted_mse(class_weights):
    def weighted_mse(gt, pred):
        # Formula: 
        # w_1*(y_1-y'_1)^2 + ... + w_100*(y_100-y'_100)^2 / sum(weights)
        return tf.keras.backend.sum(class_weights * tf.keras.backend.square(gt - pred)) / tf.keras.backend.sum(class_weights)
    return weighted_mse

def solve_me(inputs: InputsME) -> ResultsME:
    """Solve the MaxEnt problem

    Args:
        inputs (InputsME): Inputs

    Returns:
        ResultsME: Results
    """

    assert(inputs.target_cov_mat_non_zero.shape == (len(inputs.non_zero_idx_pairs),))

    # Invert cov mat to get initial guess for precision matrix
    init_prec_mat_non_zero = inputs.convert_mat_non_zero_to_inv_mat_non_zero(inputs.target_cov_mat_non_zero)
    if inputs.verbose:
        print("Prec mat initial guess for non-zero elements", init_prec_mat_non_zero)

    init_cov_mat_reconstructed_non_zero = inputs.convert_mat_non_zero_to_inv_mat_non_zero(init_prec_mat_non_zero)
    if inputs.verbose:
        print("Initial cov mat corresponding non-zero elements", init_cov_mat_reconstructed_non_zero)

    # Make layer and model
    lyr = LayerPrecToCovMat(
        n=inputs.n,
        non_zero_idx_pairs=inputs.non_zero_idx_pairs,
        non_zero_vals=init_prec_mat_non_zero
        )
    model = ModelME(inv_lyr=lyr)

    # Compile
    opt = tf.keras.optimizers.Adam(learning_rate=inputs.learning_rate)
    if inputs.use_weighted_loss:
        weights_loss = 1.0 / pow(inputs.target_cov_mat_non_zero,2)
        weights_loss = weights_loss.astype('float32')
        if inputs.verbose:
            print("Using weighted loss functing with weights:", weights_loss)
        loss_fn = custom_weighted_mse(weights_loss)
    else:
        loss_fn = tf.keras.losses.MeanSquaredError()
    model.compile(optimizer=opt,
                loss=loss_fn,
                run_eagerly=False)
    
    # Inputs outputs
    # NOTE: Do NOT use batch_size=1, use anything greater like batch_size=2
    # Make just enough for the batch size to cover
    batch_size = 2
    train_inputs = np.full(
        shape=(batch_size,1),
        fill_value=np.array([2])
        )

    train_outputs = np.full(
        shape=(batch_size,inputs.n_non_zero),
        fill_value=inputs.target_cov_mat_non_zero
        )

    # Train!
    # NOTE: Do NOT use batch_size=1, use anything greater like batch_size=2
    model.fit(
        train_inputs, 
        train_outputs, 
        epochs=inputs.epochs, 
        batch_size=batch_size
        )

    # Return solution & model
    learned_prec_mat_non_zero = model.inv_lyr.non_zero_vals.numpy()
    learned_cov_mat = convert_mat_non_zero_to_inv_mat(
        n=inputs.n,
        non_zero_idx_pairs=inputs.non_zero_idx_pairs,
        mat_non_zero=learned_prec_mat_non_zero
        )

    return ResultsME(
        inputs=inputs,
        trained_model=model,
        init_prec_mat_non_zero=init_prec_mat_non_zero,
        init_cov_mat_reconstructed_non_zero=init_cov_mat_reconstructed_non_zero,
        learned_prec_mat_non_zero=learned_prec_mat_non_zero,
        learned_cov_mat=learned_cov_mat
        )



import setuptools
from distutils.core import setup

with open("README_pypi.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='tfConstrainedGauss',
    version='0.1.0',
    packages=['tfConstrainedGauss','tfConstrainedGauss/solve_id','tfConstrainedGauss/solve_me'],
    author='Oliver K. Ernst',
    author_email='oliver.k.ernst@gmail.com',
    url='https://github.com/smrfeld/tf-constrained-gauss',
    python_requires='>=3.7',
    description="TensorFlow package for estimating constrained precision matrices",
    long_description=long_description,
    long_description_content_type="text/markdown"
    )